package it.bz.prov.mercurius.logic;

import it.bz.prov.mercurius.entity.Fornitura;
import it.bz.prov.mercurius.entity.Particella;
import it.bz.prov.mercurius.ibusiness.IBusRemote;
import it.bz.prov.mercurius.util.Constants;
import it.bz.prov.mercurius.util.Helper;
import it.bz.prov.mercurius.util.TimerInfo;
import it.bz.prov.mercurius.util.WSUtil;
import it.sian.cooperazione.schema.fornitura.InputInvioFornitura;
import it.sian.cooperazione.schema.fornitura.OutputInvioFornitura;
import it.sian.cooperazione.schema.fornitura.OutputParticelleNonCaricate;
import it.sian.cooperazione.schema.fornitura.OutputParticelleNonCaricateDatiFornituraIncongruente;
import it.sian.cooperazione.schema.fornitura.OutputParticelleNonCaricateDatiParticelleNonCaricateParticellaIncongruente;
import it.sian.cooperazione.wspdd.services.InterService;

import java.io.IOException;
import java.math.BigDecimal;
import java.rmi.RemoteException;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import javax.ejb.Stateless;
import javax.ejb.Timeout;
import javax.ejb.Timer;
import javax.ejb.TimerService;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.Query;

import org.apache.log4j.Logger;

@Stateless
public class SenderBean implements IBusRemote {

	Logger log = Logger.getLogger(this.getClass());

	@PersistenceContext(
			unitName = "SPADEV")
	private EntityManager emanGis;

	@Resource
	private TimerService timerService;

	@PostConstruct
	private void init() {
		try {
			Helper.loadProperties();

		} catch (IOException ioe) {
			System.err.println(ioe.getMessage());
		}

		return;
	}

	@Timeout
	@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
	private void send(Timer t) {
		final String methodName = "send";
		log.info(methodName);
		// TODO: Read by codiceComune, max 200 minus those already send (need
		// flag on table)

		TimerInfo tInf = (TimerInfo) t.getInfo();
		if (tInf != null && tInf.getLabel().equalsIgnoreCase(Constants.TIMER_SEND_PARTICELLE)) {
			String codComune = tInf.getElement();
			Query q = emanGis.createNamedQuery("Particella.getBySubQuery");
			q.setParameter("codiceComune", codComune);
			q.setMaxResults(Constants.MAX_REQUEST_AGEA);
			List<Particella> particelle = q.getResultList();

			if (particelle.size() > 0) {
				// make fornitura

				Transformer tr = new Transformer(codComune);
				InputInvioFornitura fornAGEA = tr.getFornitura(particelle);

				WSUtil wsu = new WSUtil();
				OutputInvioFornitura respForn = null;
				InterService is = wsu.getAllineamentoGisService(Constants.FORNITURA_PARTICELLE);
				try {
					log.info("AGEA WEBSERVICE INVOKING");
					respForn = is.fornituraParticelleVariate(fornAGEA);
					log.info("AGEA WEBSERVICE END");
				} catch (RemoteException rE) {
					log.error(rE);
				}
				if (respForn != null) {
					List<Fornitura> fornMercurius = tr.getFornitura(respForn.getMessaggioRitorno(), respForn.getDati(), new Date());
					persist(fornMercurius);
					log.info("Attention: next timeout @ " + t.getNextTimeout());
				}
				if (particelle.size() < 200) {
					t.cancel();
				}

			} else {
				log.warn("Attention: No Particelle to be send for comune: " + codComune);
				t.cancel();
			}

		}

	}

	@Timeout
	@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
	private void checkFornitura(Timer t) {
		String idFornitura = null;
		TimerInfo tI = (TimerInfo) t.getInfo();
		if (tI.getLabel().equalsIgnoreCase(Constants.TIMER_CHECK_FORNITURA)) {
			idFornitura = tI.getElement();
			Query q = emanGis.createNamedQuery("Fornitura.ByAgeaIdWithoutResponse");
			q.setParameter("idAgea", new Integer(idFornitura));
			List<Fornitura> forniture = q.getResultList();

			if (forniture.size() > 0) {
				WSUtil wsu = new WSUtil();
				InterService is = wsu.getAllineamentoGisService(Constants.PARTICELLE_NON_CARICATE);
				try {
					OutputParticelleNonCaricate outParNonCar = is.particelleNonCaricate(new BigDecimal(idFornitura));

					if (outParNonCar != null && outParNonCar.getMessaggioRitorno() != null) {
						// there is a valid response
						if (outParNonCar.getMessaggioRitorno().getCodRitorno() != null
								&& outParNonCar.getMessaggioRitorno().getMessaggio().equalsIgnoreCase(Constants.OK_AGEA.toString())) {

							if (outParNonCar.getDati() != null) {
								OutputParticelleNonCaricateDatiFornituraIncongruente fornituraKO = outParNonCar.getDati().getFornituraIncongruente();
								OutputParticelleNonCaricateDatiParticelleNonCaricateParticellaIncongruente particelleKO[] = outParNonCar.getDati()
										.getParticelleNonCaricate();

								if (fornituraKO != null) {
									this.updateFornitura(fornituraKO, Constants.FORNITURA_INCONGRUENTE, forniture);
								} else if (particelleKO != null) {
									this.updateFornitura(particelleKO, Constants.FORNITURA_INCONGRUENTE, forniture);
								} else {
									// should be: everything is OK from AGEA
									// side
									this.updateFornitura(null, Constants.OK_AGEA.toString(), forniture);
								}

							}

						} else {
							log.error(outParNonCar.getMessaggioRitorno().getCodRitorno().toString() + ".:. "
									+ outParNonCar.getMessaggioRitorno().getMessaggio());
						}
					}

				} catch (RemoteException rmi) {
					log.error(rmi);
				}

			} else {
				log.warn(".:. Fornitura " + idFornitura + " has been already checked!");
				t.cancel();
			}
		}

	}

	private void updateFornitura(Object obj, String type, List<Fornitura> forniture) {
		OutputParticelleNonCaricateDatiFornituraIncongruente fornituraKO = null;
		OutputParticelleNonCaricateDatiParticelleNonCaricateParticellaIncongruente particelleKO[] = null;

		Iterator<Fornitura> itFo = forniture.iterator();
		String idFornitura = forniture.get(0).getpKey().getId().toString();

		java.sql.Date now = new java.sql.Date(new Date().getTime());

		if (type.equalsIgnoreCase(Constants.FORNITURA_INCONGRUENTE) || type.equalsIgnoreCase(Constants.OK_AGEA.toString())) {
			// these cases need a bulk update
			Query q = emanGis.createNamedQuery("Fornitura.BulkUpdate");

			String motivo = null;
			String tipo = null;
			if (type.equalsIgnoreCase(Constants.FORNITURA_INCONGRUENTE)) {
				fornituraKO = (OutputParticelleNonCaricateDatiFornituraIncongruente) obj;
				motivo = fornituraKO.getIncongruenza().getMotivo();
				tipo = Constants.FORNITURA_INCONGRUENTE;
			} else {
				motivo = Constants.OK_AGEA.toString();
				tipo = motivo;
			}

			emanGis.flush();
			emanGis.clear();
			q.setParameter("tipo", tipo);
			q.setParameter("motivoAGEA", motivo);
			q.setParameter("update", now);
			q.setParameter("idAGEA", new Integer(idFornitura));
			int updates = q.executeUpdate();
			log.info("UPDATE for " + updates + " forniture!");

		} else if (type.equalsIgnoreCase(Constants.PARTICELLE_NON_CARICATE)) {
			for (int i = 0; i < particelleKO.length; i++) {
				OutputParticelleNonCaricateDatiParticelleNonCaricateParticellaIncongruente particellaKO = particelleKO[i];
				Fornitura f = getMatch(particellaKO, forniture);
				if (f != null) {
					f.setMotivo(particellaKO.getMotivo());
					f.setLastUpdate(now);
					f.setTipoFornitura(Constants.PARTICELLA_INCONGRUENTE);
					// TODO should work because bean is still watched!
				} else {
					log.error("Couldn't find match on fornitura!");
				}
			}
		}

	}

	private Fornitura getMatch(OutputParticelleNonCaricateDatiParticelleNonCaricateParticellaIncongruente particellaKO, List<Fornitura> forniture) {
		Iterator<Fornitura> itForn = forniture.iterator();
		boolean codComune, particella, foglio, sub;

		while (itForn.hasNext()) {
			Fornitura f = itForn.next();
			codComune = false;
			particella = false;
			foglio = false;
			sub = false;
			if (f.getCodComune().equalsIgnoreCase(particellaKO.getCodComune())) {
				codComune = true;
			}
			if (f.getParticella().equalsIgnoreCase(particellaKO.getParticella())) {
				particella = true;
			}
			if (f.getFoglio().equalsIgnoreCase(particellaKO.getFoglio().toString())) {
				foglio = true;
			}
			if ((f.getSubalterno() == null && particellaKO.getSub() == null) || (f.getSubalterno().equalsIgnoreCase(particellaKO.getSub()))) {
				sub = true;
			}

			if (codComune && particella && foglio && sub) {
				return f;
			}

		}
		return null;
	}

	private void persist(List<Fornitura> fornMercurius) {
		if (fornMercurius != null) {
			Iterator<Fornitura> itF = fornMercurius.iterator();
			while (itF.hasNext()) {
				emanGis.persist(itF.next());
			}
		}

	}

	/**
	 * returns a List of Particella defined by max results, index, and
	 * codiceComune
	 * 
	 * @param max
	 * @param index
	 * @param codiceComune
	 * @return
	 */
	private List<Particella> getParticelle(int max, int index, String codiceComune) {
		log.info("max is " + max + " index is " + index);
		Query q = emanGis.createNamedQuery("Particella.findByCodiceComune");
		q.setParameter("codiceComune", codiceComune);
		// TODO: add parameter flag "SENT" = "N" when table has been changed
		return q.setMaxResults(max).setFirstResult(index).getResultList();

	}

	private Timer getTimer(String label, String element) {
		for (Object obj : timerService.getTimers()) {
			if (obj != null) {
				Timer t = (Timer) obj;

				TimerInfo tInfo = (TimerInfo) t.getInfo();
				if (tInfo != null && tInfo.getLabel().equalsIgnoreCase(label) && tInfo.getElement().equalsIgnoreCase(element)) {
					return t;

				}

			}

		}
		return null;
	}

	@Override
	public Date checkTransmission(String idFornitura) {
		Date date = buildTimer(Constants.TIMER_CHECK_FORNITURA, idFornitura, 5);
		// Timer t = null;
		// t = getTimer(Constants.TIMER_CHECK_FORNITURA, idFornitura);
		// if (t != null) {
		// return t.getNextTimeout();
		// } else {
		// Date now = new Date();
		// long ldate = now.getTime();
		// // schedule for first start
		// ldate = ldate + (1 * 60 * 1000);
		// TimerInfo tInfo = new TimerInfo();
		// tInfo.setLabel(Constants.TIMER_CHECK_FORNITURA);
		// tInfo.setElement(idFornitura);
		// t = timerService.createTimer(new Date(ldate), 20 * 60 * 1000, tInfo);
		// }

		return date;
	}

	@Override
	public Date sendGIS(String codiceComune) {

		Date date = buildTimer(Constants.TIMER_SEND_PARTICELLE, codiceComune, 5);
		//
		// Timer t = null;
		// t = getTimer(Constants.TIMER_SEND_PARTICELLE, codiceComune);
		// if (t != null) {
		// return t.getNextTimeout();
		// } else {
		// Date now = new Date();
		// long ldate = now.getTime();
		// // schedule for first start
		// ldate = ldate + (1 * 60 * 1000);
		// TimerInfo tInfo = new TimerInfo();
		// tInfo.setLabel(Constants.TIMER_SEND_PARTICELLE);
		// tInfo.setElement(codiceComune);
		// t = timerService.createTimer(new Date(ldate), 5 * 60 * 1000, tInfo);
		// }

		return date;

	}

//	@Override
//	public void deleteTimers() {
//		int i = 0;
//		for (Object obj : timerService.getTimers()) {
//			if (obj != null) {
//				Timer t = (Timer) obj;
//				t.cancel();
//				i++;
//			}
//		}
//		log.info("Deleted " + i + " timers!");
//
//	}

	private Date buildTimer(String label, String element, int minutes) {
		Timer timer = null;
		if (label != null && element != null) {
			timer = getTimer(label, element);
			if (timer == null) {
				TimerInfo tInfo = new TimerInfo(label, element);
				long interval = minutes * 60 * 1000;
				timer = timerService.createTimer(new Date(), interval, tInfo);
				log.info("PI* Created Timer with label:"+label+" and element: "+element+". Next timeout @ " + timer.getNextTimeout());

			} else {
				log.info("PI* Found timer with label:"+label+" and element:"+element+ "and next timeout @ "+timer.getNextTimeout());
			}
		}

		return timer.getNextTimeout();
	}

}
